// base.cpp

#ifndef _CONSOLE
#include "stdafx.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <malloc.h>

#include "base.h"

/////////////////////////////////////////////////////////////
// Vector functions by taka given from C Magagine appendix
/////////////////////////////////////////////////////////////

#define InnerProduct3D(v1, v2)	( (v1)->x*(v2)->x + (v1)->y*(v2)->y + (v1)->z*(v2)->z)

// added by okada on 98.08.04
Vector *RandomVector()
{
  static Vector v;

  v.x = (double)rand() / 32768.0; 
  v.y = (double)rand() / 32768.0;
  v.z = (double)rand() / 32768.0;
  return &v;
}

double Length3D(Vector  *v)
{
  double InP = InnerProduct3D(v,v);
  if ( InP > F_TH ) return sqrt(InP);
  else return 0.0;
}

double DistanceBetween(Vector  *v1, Vector  *v2)
{
  Vector v;
  VSub(&v, v1, v2);
  return Length3D( &v );
}

void VNormalize(Vector  *v)
{
    double l = Length3D(v);
    if ( l > F_TH ) VDiv( v, v, l );
}

void VInitialize(Vector *v, double d)
{
  v->x = v->y = v->z = d;
}

#define Between(begin, end, x)	(( (begin) < (x) && (x) <( end) ) || ( (begin) > (x) && (x) > (end) ))

void VAdd(Vector  *sum, Vector  *v1, Vector  *v2)
{
    sum->x = v1->x + v2->x;
    sum->y = v1->y + v2->y;
    sum->z = v1->z + v2->z;
}

void VSub(Vector  *sub, Vector  *v1, Vector  *v2)
{
    sub->x = v1->x - v2->x;
    sub->y = v1->y - v2->y;
    sub->z = v1->z - v2->z;
}

void VDiv(Vector  *v1, Vector  *v0, Vector *v2)
{
    v1->x = v0->x / v2->x;
    v1->y = v0->y / v2->y;
    v1->z = v0->z / v2->z;
}

void VMul(Vector  *v1, Vector  *v0, Vector *v2)
{
    v1->x = v0->x * v2->x;
    v1->y = v0->y * v2->y;
    v1->z = v0->z * v2->z;
}

void VAdd(Vector  *v1, Vector  *v0, double f)
{
    v1->x = v0->x + f;
    v1->y = v0->y + f;
    v1->z = v0->z + f;
}

void VSub(Vector  *v1, Vector  *v0, double f)
{
    v1->x = v0->x - f;
    v1->y = v0->y - f;
    v1->z = v0->z = f;
}

void VDiv(Vector  *v1, Vector  *v0, double f)
{
    v1->x = v0->x / f;
    v1->y = v0->y / f;
    v1->z = v0->z / f;
}

void VMul(Vector  *v1, Vector  *v0, double f)
{
    v1->x = v0->x * f;
    v1->y = v0->y * f;
    v1->z = v0->z * f;
}


double VInner(Vector  *v1, Vector  *v2)
{
    return v1->x * v2->x + v1->y * v2->y + v1->z * v2->z;
}

void VVector(Vector  *x, Vector  *v0, Vector  *v1)
{
    Vector v;
    v.x = v0->y * v1->z - v0->z * v1->y;
    v.y = v0->z * v1->x - v0->x * v1->z;
    v.z = v0->x * v1->y - v0->y * v1->x;
    *x = v;
}

BOOLEAN VEqual( Vector *v1, Vector *v2 )
{
  return ( FEQUAL( v1->x, v2->x ) && FEQUAL( v1->y, v2->y ) && FEQUAL( v1->z, v2->z ) );
}

BOOLEAN EqualVector(Vector *v1, Vector *v2)
{
  return ( FEQUAL( Length3D(v1), Length3D(v2) ) );
}

BOOLEAN SmallerVector(Vector *v1, Vector *v2)
{
  return ( Length3D(v1) < Length3D(v2) );
}

BOOLEAN GreaterVector(Vector *v1, Vector *v2)
{
  return ( Length3D(v1) > Length3D(v2) );
} 

BOOLEAN EqualVector(Vector *v1, double f)
{
  return ( FEQUAL( Length3D(v1), f ) );
}

BOOLEAN SmallerVector(Vector *v1, double f)
{
  return ( Length3D(v1) < f );
}

BOOLEAN GreaterVector(Vector *v1, double f)
{
  return ( Length3D(v1) > f );
}

// MuraMura
double VAngle(Vector *fromVec, Vector *toVec)
{
  return( acos( VInner(fromVec, toVec) / ( Length3D(fromVec) * Length3D(toVec) ) ) );
}

//////////////////////////////////
//  TranslationMatrix routines
//////////////////////////////////

// c = a x b
void MatrixMatrixMatrix(Matrix44 c, Matrix44 a, Matrix44 b)
{
    int i, j;
    Matrix44 c0;
    for (i = 0; i < 3; i++) {
	for (j = 0; j < 3; j++) {
	    c0[i][j] = a[i][0]*b[0][j] + a[i][1]*b[1][j] + a[i][2]*b[2][j];
	}
	c0[3][i] = a[3][0]*b[0][i] + a[3][1]*b[1][i] + a[3][2]*b[2][i] + b[3][i];
    }
    memcpy(c, c0, sizeof(Matrix44));	// c := c0
}

// E = a x b

// the routine given from Takahasi-kun

#define MATRIX_ARRAY_LENGTH 16
#define MATRIX_COLUMNS 4
#define MATRIX_ROWS 4

#define ELM(m,i,j)  ((m)[(i)*MATRIX_COLUMNS+(j)])

#define MATRIX_NORMAL 0
#define MATRIX_SINGULAR -1

static int matrix_invert (double *dst, double *m)
{
 // normally dst is required as unit.
 int i, j, k;

 // gauss elimination method

 // LU decomposition
 for (i = 0; i < MATRIX_ROWS; i++){
 // find pivot
   for (j = i; j < MATRIX_ROWS && FEQUAL( ELM(m,j,i), 0.0 ); j++);
   if (j == MATRIX_ROWS) return MATRIX_SINGULAR;
   if (i != j){
 // swap rows and normalize
     double p = ELM(m,j,i);
     double t;
     for (k = i; k < MATRIX_COLUMNS; k++){
       t = ELM(m,j,k);
       ELM(m,j,k) = ELM(m,i,k); ELM(m,i,k) = t / p;
     }
     for (k = 0; k < MATRIX_COLUMNS; k++){
       t = ELM(dst,j,k);
       ELM(dst,j,k) = ELM(dst,i,k); ELM(dst,i,k) = t / p;
     }
   } else {
 // must be normalize?
     double p = ELM(m,j,i);
     if ( !FEQUAL( p, 1.0) ){
 // normalize
       for (k = i; k < MATRIX_COLUMNS; k++) ELM(m,i,k) /= p;
       for (k = 0; k < MATRIX_COLUMNS; k++) ELM(dst,i,k) /= p;
     }
   }
   for (j = i+1; j < MATRIX_ROWS; j++){
     double t = ELM(m,j,i);
     for (k = i; k < MATRIX_COLUMNS; k++) ELM(m,j,k) -= ELM(m,i,k) * t;
     for (k = 0; k < MATRIX_COLUMNS; k++) ELM(dst,j,k) -= ELM(dst,i,k) * t;
   }
 }

 // backward assignment
 for (i = MATRIX_ROWS - 1; i >= 0; i--)
   for (j = i-1; j >= 0; j--){
     double t = ELM(m,j,i);
     for (k = 0; k < MATRIX_COLUMNS; k++) ELM(dst,j,k) -= ELM(dst,i,k) * t;
   }

  return MATRIX_NORMAL;
}

void InverseMatrix(Matrix44 b, Matrix44 a)
{
  int i,j;
  double matA[MATRIX_ARRAY_LENGTH],matB[MATRIX_ARRAY_LENGTH];
  for(i=0; i < MATRIX_COLUMNS; i++){
    for(j=0; j < MATRIX_ROWS -1; j++){
      ELM(matA, j, i) = a[i][j];
    }
    ELM(matA, 3, i) = 0.0;
  }
  ELM(matA, 3, 3) = 1.0;

  for(i=0; i < MATRIX_COLUMNS; i++)
    for(j=0; j < MATRIX_ROWS; j++){
      if ( i == j ) ELM(matB, j, i) = 1;
      else ELM(matB, j, i) = 0.0;
  }
 
  matrix_invert(matB, matA);

  for(i=0; i < MATRIX_COLUMNS; i++)
    for(j=0; j < MATRIX_ROWS -1; j++) b[i][j] = ELM(matB, j, i);
}

// a = E
void InitializeMatrix(Matrix44 a)
{
  a[0][0] = 1.0; a[0][1] = 0.0; a[0][2] = 0.0; a[0][3] = 0.0;
  a[1][0] = 0.0; a[1][1] = 1.0; a[1][2] = 0.0; a[1][3] = 0.0;
  a[2][0] = 0.0; a[2][1] = 0.0; a[2][2] = 1.0; a[2][3] = 0.0;
  a[3][0] = 0.0; a[3][1] = 0.0; a[3][2] = 0.0; a[3][3] = 1.0;
}

void InitializeMatrix(Matrix44 a, double d)
{
  for(unsigned int i = 0; i < 4; i++)
    for(unsigned int j = 0; j < 4; j++) a[i][j] = d;
}
           
// a = a x b
void MatrixMatrix(Matrix44 a, Matrix44 b)
{
  MatrixMatrixMatrix(a, a, b);  
}

// _v = v0 x mat
void VectorMatrix(Vector  *_v, Vector *v, Matrix44 mat)
{
  Vector v1;
  v1.x = v->x * mat[0][0] + v->y * mat[1][0] + v->z * mat[2][0] + mat[3][0];
  v1.y = v->x * mat[0][1] + v->y * mat[1][1] + v->z * mat[2][1] + mat[3][1];
  v1.z = v->x * mat[0][2] + v->y * mat[1][2] + v->z * mat[2][2] + mat[3][2];
  *_v = v1;
}

// a = b
void CopyMatrix(Matrix44 a, Matrix44 b)
{
  memcpy(a, b, sizeof(Matrix44)); 
}

void CopyMatrixWithoutPosition(Matrix44 a, Matrix44 b)
{
  Vector v1;
  v1.x = a[3][0]; v1.y = a[3][1]; v1.z = a[3][2];
  memcpy(a, b, sizeof(Matrix44));
  a[3][0] = v1.x; a[3][1] = v1.y; a[3][2] = v1.z;
}

// a == b
BOOLEAN EqualMatrix(Matrix44 a, Matrix44 b)
{
  return ( FEQUAL( a[0][0], b[0][0] ) && FEQUAL( a[0][1], b[0][1] ) && FEQUAL( a[0][2], b[0][2] ) &&
           FEQUAL( a[1][0], b[1][0] ) && FEQUAL( a[1][1], b[1][1] ) && FEQUAL( a[1][2], b[1][2] ) &&
           FEQUAL( a[2][0], b[2][0] ) && FEQUAL( a[2][1], b[2][1] ) && FEQUAL( a[2][2], b[2][2] ) &&
           FEQUAL( a[3][0], b[3][0] ) && FEQUAL( a[3][1], b[3][1] ) && FEQUAL( a[3][2], b[3][2] ) );
}

// LocalTranslate(tx, ty, tz)
void LocalTranslateMatrix(Matrix44 mat, double tx, double ty, double tz)
{
  Matrix44 tmat;
  InitializeMatrix(tmat);
  tmat[3][0] = tx;
  tmat[3][1] = ty;
  tmat[3][2] = tz;
  MatrixMatrixMatrix(mat, tmat, mat);
}

// GlobalTranslate(tx, ty, tz)
void GlobalTranslateMatrix(Matrix44 mat, double tx, double ty, double tz)
{
  Matrix44 tmat;
  InitializeMatrix(tmat);
  tmat[3][0] = tx;
  tmat[3][1] = ty;
  tmat[3][2] = tz;
  MatrixMatrixMatrix(mat, mat, tmat);
}

// LocalRotateMatrix(axis, theta)
void LocalRotateMatrix(Matrix44 mat, Vector  *axis, double theta)
{
  Matrix44 tmat;
  MakeRotateMatrix(tmat, axis, theta);
  MatrixMatrixMatrix(mat, tmat, mat);
}

// GlobalRotateMatrix(axis, theta)
void GlobalRotateMatrix(Matrix44 mat, Vector  *axis, double theta)
{
  Matrix44 tmat;
  MakeRotateMatrix(tmat, axis, theta);
  MatrixMatrixMatrix(mat, mat, tmat);
}

// LocalScale(sx, sy, sz)
void LocalScaleMatrix(Matrix44 mat, double sx, double sy, double sz)
{
  Matrix44 tmat;
  InitializeMatrix(tmat);
  tmat[0][0] *= sx;
  tmat[1][1] *= sy;
  tmat[2][2] *= sz;
  MatrixMatrixMatrix(mat, tmat, mat);
}

// GlobalScale(sx, sy, sz)
void GlobalScaleMatrix(Matrix44 mat, double sx, double sy, double sz)
{
  Matrix44 tmat;
  InitializeMatrix(tmat);
  tmat[0][0] *= sx;
  tmat[1][1] *= sy;
  tmat[2][2] *= sz;
  MatrixMatrixMatrix(mat, mat, tmat);
}

// Scan position vector from Matrix

Vector *PositionMatrix(Matrix44 mat)
{
  static Vector v;
  v.x = mat[3][0]; v.y = mat[3][1]; v.z = mat[3][2];
  return &v;
}

void SetPositionMatrix(Matrix44 mat, Vector *v)
{
  mat[3][0] = v->x; mat[3][1] = v->y; mat[3][2] = v->z;
}

// added by okada on 18/08/2000
// Scan Euler angle vector from Matrix
Vector *AngleMatrix(Matrix44 mat)
{
	static Vector v;

	if ( mat[0][0] == 0 && mat[0][1] == 0 ){
		v.x = atan2( mat[1][0], mat[1][1] );
		v.y = PIHALF;
		v.z = 0;
	} else {
		v.x = atan2( mat[1][2], mat[2][2] );
		v.y = asin( - mat[0][2] );
		v.z = atan2( mat[0][1], mat[0][0] );
	}
	return &v;
}

double DistanceBetween(Matrix44 a, Matrix44 b)
{
  static Vector v1, v2;
  v1 = *PositionMatrix( a );
  v2 = *PositionMatrix( b );
  return DistanceBetween( &v1, &v2 );
}

void ClearMatrix(Matrix44 mat)
{
  for(int i=0; i<4; i++)
    for(int j=0; j<3; j++) mat[i][j] = 0.0;
}

void AddMatrix(Matrix44 matA, Matrix44 matB, double d)
{
  for(int i=0; i<4; i++)
    for(int j=0; j<3; j++) matA[i][j] = matB[i][j] + d;
}

void SubMatrix(Matrix44 matA, Matrix44 matB, double d)
{
  for(int i=0; i<4; i++)
    for(int j=0; j<3; j++) matA[i][j] = matB[i][j] - d;
}

void MulMatrix(Matrix44 matA, Matrix44 matB, double d)
{
  for(int i=0; i<4; i++)
    for(int j=0; j<3; j++) matA[i][j] = matB[i][j] * d;
}

void DivMatrix(Matrix44 matA, Matrix44 matB, double d)
{
  for(int i=0; i<4; i++)
    for(int j=0; j<3; j++) matA[i][j] = matB[i][j] / d;
}

void MatrixAddMatrix(Matrix44 matA, Matrix44 matB, Matrix44 matC)
{
  for(int i=0; i<4; i++)
    for(int j=0; j<3; j++) matA[i][j] = matB[i][j] + matC[i][j];
}

void MatrixSubMatrix(Matrix44 matA, Matrix44 matB, Matrix44 matC)
{
  for(int i=0; i<4; i++)
    for(int j=0; j<3; j++) matA[i][j] = matB[i][j] - matC[i][j];
}

void MatrixMulMatrix(Matrix44 matA, Matrix44 matB, Matrix44 matC)
{
  for(int i=0; i<4; i++)
    for(int j=0; j<3; j++) matA[i][j] = matB[i][j] * matC[i][j];
}

void MatrixDivMatrix(Matrix44 matA, Matrix44 matB, Matrix44 matC)
{
  for(int i=0; i<4; i++)
    for(int j=0; j<3; j++) matA[i][j] = matB[i][j] / matC[i][j];
}

void WriteMatrix( char *str, Matrix44 m )
{
  sprintf(str,"%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",
                     m[0][0], m[0][1], m[0][2],
                     m[1][0], m[1][1], m[1][2],
                     m[2][0], m[2][1], m[2][2],
                     m[3][0], m[3][1], m[3][2]);
}

void ReadMatrix( char *str, Matrix44 m )
{
  sscanf(str,"%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
                     &m[0][0], &m[0][1], &m[0][2],
                     &m[1][0], &m[1][1], &m[1][2],
                     &m[2][0], &m[2][1], &m[2][2],
                     &m[3][0], &m[3][1], &m[3][2]);
}

void MakeScaleMatrix(Matrix44 mat, double sx, double sy, double sz)
{
  mat[0][0] = sx;   mat[0][1] = 0.0;  mat[0][2] = 0.0;
  mat[1][0] = 0.0;  mat[1][1] = sy;   mat[1][2] = 0.0;
  mat[2][0] = 0.0;  mat[2][1] = 0.0;  mat[2][2] = sz;
  mat[3][0] = 0.0;  mat[3][1] = 0.0;  mat[3][2] = 0.0;
}

// modified by okada on 98.09.03 
void MakeRotateXMatrix(Matrix44 mat, double thea)
{
  double sx = sin(thea), cx = cos(thea);

  mat[0][0] = 1.0; mat[0][1] = 0.0; mat[0][2] = 0.0;
  mat[1][0] = 0.0; mat[1][1] = cx ; mat[1][2] = sx;
  mat[2][0] = 0.0; mat[2][1] = -sx; mat[2][2] = cx;
  mat[3][0] = 0.0; mat[3][1] = 0.0; mat[3][2] = 0.0;
}

void MakeRotateYMatrix(Matrix44 mat, double thea)
{
  double sy = sin(thea), cy = cos(thea);

  mat[0][0] = cy;  mat[0][1] = 0.0; mat[0][2] = -sy;
  mat[1][0] = 0.0; mat[1][1] = 1.0; mat[1][2] = 0.0;
  mat[2][0] = sy;  mat[2][1] = 0.0; mat[2][2] = cy;
  mat[3][0] = 0.0; mat[3][1] = 0.0; mat[3][2] = 0.0;
}

void MakeRotateZMatrix(Matrix44 mat, double thea)
{
  double sz = sin(thea), cz = cos(thea);

  mat[0][0] = cz;  mat[0][1] = sz;  mat[0][2] = 0.0;
  mat[1][0] = -sz; mat[1][1] = cz;  mat[1][2] = 0.0;
  mat[2][0] = 0.0; mat[2][1] = 0.0; mat[2][2] = 1.0;
  mat[3][0] = 0.0; mat[3][1] = 0.0; mat[3][2] = 0.0;
}

// added by okada on 98.09.03
// modified by okada on 18/08/2000 Rxyz = Rx * Ry * Rz
void MakeRotateXYZMatrix(Matrix44 mat, double x, double y, double z)
{
  double sx = sin(x), cx = cos(x),
         sy = sin(y), cy = cos(y),
         sz = sin(z), cz = cos(z);

//  mat[0][0] = cy * cz;                mat[0][1] = cy * sz;                mat[0][2] = - sz;

  mat[0][0] = cy * cz;                mat[0][1] = cy * sz;                mat[0][2] = - sy;  
  mat[1][0] = sx * sy * cz - cx * sz; mat[1][1] = sx * sy * sz + cx * cz; mat[1][2] = sx * cy;
  mat[2][0] = cx * sy * cz + sx * sz; mat[2][1] = cx * sy * sz - sx * cz; mat[2][2] = cx * cy;
  mat[3][0] = 0.0;                    mat[3][1] = 0.0;                    mat[3][2] = 0.0;
}

void MakeRotateMatrix(Matrix44 R, Vector *axis, double theta)
{
    double  ntheta = - theta,
    l = Length3D(axis),
    ct = cos(ntheta),
    rct = 1-ct,
    st = sin(ntheta),
    u = axis->x / l,
    v = axis->y / l,
    w = axis->z / l;

    R[3][0] = R[3][1] = R[3][2] = 0.0;
    R[0][0]= ct    + u*u*rct; R[0][1]= -w*st + v*u*rct; R[0][2]= v*st  + w*u*rct;
    R[1][0]= w*st  + u*v*rct; R[1][1]= ct    + v*v*rct; R[1][2]= -u*st + w*v*rct;
    R[2][0]= -v*st + u*w*rct; R[2][1]= u*st  + v*w*rct; R[2][2]= ct    + w*w*rct;
}

// added by okada at 98.05.02
// modified by okada on 98.09.03
void RotateXYZMatrix(Matrix44 R, Vector *angle)
{
  Matrix44 mat;
  /*
  MakeRotateXMatrix( mat, angle->x );
  MatrixMatrix( R, mat );
  MakeRotateYMatrix( mat, angle->y );
  MatrixMatrix( R, mat );
  MakeRotateZMatrix( mat, angle->z );
  */
  MakeRotateXYZMatrix( mat, angle->x, angle->y, angle->z );
  MatrixMatrix( R, mat );
}

// added on 98.11.19 begin
void MatrixRotateAxisInternal(Matrix44 mat, Vector *axis, double theta)
{
  Vector ov;
  ov.x = mat[3][0];
  ov.y = mat[3][1];
  ov.z = mat[3][2];

  GlobalTranslateMatrix(mat, -ov.x, -ov.y, -ov.z);
  GlobalRotateMatrix(mat, axis, theta);
  GlobalTranslateMatrix(mat, ov.x, ov.y, ov.z);
}

void MatrixRotateToInternal(Matrix44 mat0, Matrix44 mat1, Vector *retv1, double *retd1, Vector *retv2, double *retd2)
{
  Matrix44 imat;
  Vector pos, pos1, v,
         v0 = {0.0, 0.0, 0.0}, v1 = { 0.0, 0.0, 1.0 }, v2 = { 0.0, 0.0, 1.0 };
  pos = *PositionMatrix( mat0 );
  VectorMatrix( &v1, &v1, mat0 );
  VSub(&v1, &v1, &pos);
  VNormalize(&v1);
  pos1 = *PositionMatrix( mat1 );
  VectorMatrix( &v2, &v2, mat1 );
  VSub(&v2, &v2, &pos1);
  VNormalize(&v2);
  double vin = VInner(&v1, &v2);

  InitializeMatrix(imat);
  VectorMatrix(&v0, &v0, imat);

  if ( FEQUAL( vin, 1.0 ) ){
    if ( retv1 ){
      *retv1 = v1;
      *retd1 = 0.0;
    }
  }
  else
  if ( FEQUAL( vin, -1.0 ) ){
      Vector vv1={ 1.0, 0.0, 0.0 };
      Vector vv2={ 0.0, 1.0, 0.0 };
      if ( FEQUAL ( VInner( &v1, &vv1 ), 1.0 ) ) VVector(&v, &v1, &vv2);
      else VVector(&v, &v1, &vv1);

      VectorMatrix(&v, &v, imat);
      VSub(&v,&v,&v0);
      MatrixRotateAxisInternal(mat0, &v, PI);
      if ( retv1 ){
	*retv1 = v;
        *retd1 = PI;
      }
  } else {
      VVector(&v, &v1, &v2);

      VectorMatrix(&v, &v, imat);
      VSub(&v,&v,&v0);
      MatrixRotateAxisInternal(mat0, &v, acos( vin ) );
      if ( retv1 ){
        *retv1 = v;
        *retd1 = acos( vin );
      }
  }

  v1.x = 0.0; v1.y = 1.0; v1.z = 0.0;
  v2.x = 0.0; v2.y = 1.0; v2.z = 0.0;
  VectorMatrix( &v1, &v1, mat0 );
  VSub(&v1, &v1, &pos);
  VNormalize(&v1);
  VectorMatrix( &v2, &v2, mat1 );
  VSub(&v2, &v2, &pos1);
  VNormalize(&v2);
  vin = VInner(&v1, &v2);
  if ( FEQUAL( vin, 1.0 ) ){
    if ( retv2 ){
      *retv2 = v1;
      *retd2 = 0.0;
    }
  }
  else
  if ( FEQUAL( vin, -1.0 ) ){
      Vector vv1={ 1.0, 0.0, 0.0 };
      Vector vv2={ 0.0, 1.0, 0.0 };
      if ( FEQUAL ( VInner( &v1, &vv1 ), 1.0 ) ) VVector(&v, &v1, &vv2);
      else VVector(&v, &v1, &vv1);

      VectorMatrix(&v, &v, imat);
      VSub(&v,&v,&v0);
      MatrixRotateAxisInternal(mat0, &v, PI);
      if ( retv2 ){
	*retv2 = v;
	*retd2 = PI;
      }
  } else {
      VVector(&v, &v1, &v2);

      VectorMatrix(&v, &v, imat);
      VSub(&v,&v,&v0);
      MatrixRotateAxisInternal(mat0, &v, acos( vin ) );
      if ( retv2 ){
	*retv2 = v;
	*retd2 = acos( vin );
      }
  }
}

void CopyMatrixWithoutScaling(Matrix44 mat2, Matrix44 mat1)
{
  InitializeMatrix(mat2);
  MatrixRotateToInternal(mat2, mat1);
  mat2[3][0] = mat1[3][0];
  mat2[3][1] = mat1[3][1];
  mat2[3][2] = mat1[3][2];
}

//  followings are new
void TranslateMatrix(Matrix44 mat, Vector *v)
{
	SetPositionMatrix( mat, v );
}

void MakeRotateZXYMatrix(Matrix44 mat, double x, double y, double z)
{
  double sx = sin(x), cx = cos(x),
         sy = sin(y), cy = cos(y),
         sz = sin(z), cz = cos(z);
  mat[0][0] = cy*cz - sz*sx*sy; mat[0][1] = sz*cz+cz*sx*sy; mat[0][2] = -cx*sy;
  mat[1][0] = -sz*cx; mat[1][1] = cz*cx; mat[1][2] = sx;
  mat[2][0] = cz*sy+sz*sx*cy; mat[2][1] = sz*sy-cz*sx*cy; mat[2][2] = cx*cy;
  mat[3][0] = 0; mat[3][1] = 0; mat[3][2] = 0;
}

void RotateZXYMatrix(Matrix44 R, Vector *v)
{
	Matrix44 mat;
	MakeRotateZXYMatrix(mat, v->x, v->y, v->z);
	MatrixMatrix(R, mat);
}
